/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.osgicdi.impl;

import org.glassfish.osgijavaeebase.Extender;
import org.jboss.weld.bootstrap.WeldBootstrap;
import org.jboss.weld.bootstrap.api.Environments;
import org.jboss.weld.bootstrap.spi.BeanDeploymentArchive;
import org.jboss.weld.manager.api.WeldManager;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.BundleTracker;
import org.osgi.util.tracker.BundleTrackerCustomizer;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Logger;

import javax.enterprise.inject.Instance;

/**
 * An extender that registers a BundleTracker to trigger Weld SPI directly
 * Pl.see http://java.net/jira/browse/GLASSFISH-19215
 *
 * @author Tang Yong(tangyong@cn.fujitsu.com)
 */
public class OSGiCDIExtender implements Extender{
	
    private static final Logger logger =
            Logger.getLogger(OSGiCDIExtender.class.getPackage().getName());
    private BundleContext context;
    private BundleTracker tracker;
    private HashMap<Long, WeldContainerInfo> managedBootstraps = new HashMap<Long, WeldContainerInfo>();

    public OSGiCDIExtender(BundleContext context) {
        this.context = context;
    }

    public synchronized void start() {
    	//executorService = Executors.newSingleThreadExecutor();
    	tracker = new BundleTracker(context, Bundle.ACTIVE | Bundle.STARTING | Bundle.STOPPING, new BeanBundleTrackerCustomizer());
        tracker.open();
    }

    public synchronized void stop() {   	
    	
    	if (tracker != null) tracker.close();
        tracker = null;
        
        for(Long bundleId : managedBootstraps.keySet()){
        	WeldBootstrap bootstrap = managedBootstraps.get(bundleId).getBootstrap();
        	if (bootstrap != null){
        		ClassLoader oldTCC = Thread.currentThread().getContextClassLoader();
				try{
					ClassLoader newTCC = managedBootstraps.get(bundleId).getClassLoader();
					Thread.currentThread().setContextClassLoader(newTCC);
					bootstrap.shutdown();
				}catch(RuntimeException e){
					e.printStackTrace();
				}finally{
					Thread.currentThread().setContextClassLoader(oldTCC);
				}								
        	}
        	
        	managedBootstraps.get(bundleId).getInstanceSR().unregister();
        }   
        
        managedBootstraps.clear();
        managedBootstraps = null;        
    }
    
    private static class TCCLClassLoader extends ClassLoader {
        private final Bundle bundle;

        private final ClassLoader infra;

        public TCCLClassLoader(Bundle bundle, ClassLoader infraClassLoader) {
            this.bundle = bundle;
            this.infra = infraClassLoader;
        }

        @Override
        public Class<?> loadClass(String name) throws ClassNotFoundException {
            Class<?> loadedClass = null;
            try {
                loadedClass = bundle.loadClass(name);
            } catch (ClassNotFoundException cnfe) {
                loadedClass = infra.loadClass(name);
            }
            return loadedClass;
        }

        @SuppressWarnings("unchecked")
		@Override
        public Enumeration<URL> getResources(String s) throws IOException {
            Set<URL> urls = new HashSet<URL>();
            Enumeration<?> enumBundle = bundle.getResources(s);
            Enumeration<?> enumInfra = infra.getResources(s);
            if (enumBundle != null) {
                urls.addAll((Collection<? extends URL>) Collections.list(enumBundle));
            }
            if (enumInfra != null) {
                urls.addAll((Collection<? extends URL>) Collections.list(enumInfra));
            }
            return Collections.enumeration(urls);
        }
    }
    
    private class WeldContainerInfo{
    	private WeldBootstrap bootstrap;
    	private ClassLoader cl;
    	private ServiceRegistration instanceSR;
    	
    	public WeldContainerInfo(WeldBootstrap bootstrap, ClassLoader cl, ServiceRegistration instanceSR){
    		this.bootstrap = bootstrap;
    		this.cl = cl;
    		this.instanceSR = instanceSR;
    	}
    	
    	public WeldBootstrap getBootstrap(){
    		return bootstrap;
    	}
    	
    	public ClassLoader getClassLoader(){
    		return cl;
    	}
    	
    	public ServiceRegistration getInstanceSR(){
    		return instanceSR;
    	}
    }
    
    private class BeanBundleTrackerCustomizer implements BundleTrackerCustomizer {

		@Override
		public Object addingBundle(final Bundle bundle, BundleEvent event) {			
			if ( isPlainOSGiBundle(bundle) ){
				if ( isGlassfishCDIEnabled(bundle) && isCDIBundle(bundle) ){
					final int state = bundle.getState();
					if ( isReady(event, state) ) {
						startWeldIntegration();
						WeldBootstrap bootstrap = new WeldBootstrap();
    		            OSGiBeanDeploymentArchiveFactory factory = new OSGiBeanDeploymentArchiveFactory();
    		            OSGiBeanDeployment deployment = new OSGiBeanDeployment(bundle, bootstrap, factory);
    		            BeanDeploymentArchive beanDeploymentArchive = deployment.getBeanDeploymentArchive();
    		            
    		            ClassLoader oldTCC = Thread.currentThread().getContextClassLoader();   					
    					try{
    						ClassLoader newTCC = new TCCLClassLoader(bundle, this.getClass().getClassLoader());
    						Thread.currentThread().setContextClassLoader(newTCC);
    						
    						bootstrap.startContainer(Environments.SE, deployment);
        		            // Start the container
    						bootstrap.startInitialization();
    						bootstrap.deployBeans();  
    						
    						//Register Instance<Object> as OSGi Service with ServiceProperty BundleId
    						WeldManager manager = bootstrap.getManager(beanDeploymentArchive);
					        Instance<Object> instance = manager.instance();
					        Properties properties = new Properties();
					        Long bundleId = Long.valueOf(bundle.getBundleId());
           				    properties.put("BundleID", bundleId);
           				    ServiceRegistration sr = context.registerService(Instance.class.getName(), instance, properties);
    						
    						bootstrap.validateBeans();
    						bootstrap.endInitialization();	
        		            
    						WeldContainerInfo wci = new WeldContainerInfo(bootstrap, newTCC, sr);
        		            managedBootstraps.put(bundleId, wci);
        		            
    					}catch(RuntimeException e){
    						e.printStackTrace();
    					}finally{
    						Thread.currentThread().setContextClassLoader(oldTCC);
    					}
																				
		                return bundle;
		            }
				}
			}
			
            return null;
		}
		
		private void startWeldIntegration(){
			Bundle[] installedBundles = context.getBundles();
			for(Bundle bundle : installedBundles){
				if (Constants.WEDINTEGRATION_SYMNAME.equals(bundle.getSymbolicName())){
					if (bundle.getState() != Bundle.ACTIVE){
						try {
							System.out.println("Begin to invocating bundle.start method");
							//bundle.start(Bundle.START_TRANSIENT);
							//Temp Fixing way, please seeing http://java.net/jira/browse/GLASSFISH-19326
							//And https://issues.apache.org/jira/browse/FELIX-3713
							//Once FELIX-3713 is fixed, bundle.start(Bundle.START_TRANSIENT) will be used
							bundle.start();
						} catch (BundleException e) {
							throw new RuntimeException(e);
						}
					}
				}
			}
		}

		private boolean isGlassfishCDIEnabled(Bundle bundle){
			Object gfcdiFlag = bundle.getHeaders().get(Constants.GLASSFISH_CDIENABLED);
			if (gfcdiFlag != null && "true".equals(gfcdiFlag)) {
		            return true;
		    }
			
			return false;
		}
		
		private boolean isPlainOSGiBundle(Bundle bundle){
			final Dictionary<?, ?> headers = bundle.getHeaders();
			
			return headers.get(Constants.EXPORT_EJB) == null && 
	        		headers.get(Constants.WEB_CONTEXT_PATH) == null && 
	                  headers.get(org.osgi.framework.Constants.FRAGMENT_HOST) == null;
		}
		
		private boolean isCDIBundle(Bundle bundle) {
	        URL url = bundle.getEntry(Constants.BEANS_XML_PATH);
	        return url != null;
		}

		private boolean isReady(BundleEvent event, int state) {
			return state == Bundle.ACTIVE ||
                    (state == Bundle.STARTING && (event != null && event.getType() == BundleEvent.LAZY_ACTIVATION));
		}

		@Override
		public void modifiedBundle(Bundle bundle, BundleEvent event,
				Object object) {
			if ( BundleEvent.STOPPING == event.getType() ){
				Long bundleId = Long.valueOf(bundle.getBundleId());
				
				if ( managedBootstraps.containsKey(bundleId) ){
					//Shutting Down Weld Container which belongs to current bundle
					ClassLoader oldTCC = Thread.currentThread().getContextClassLoader();
					try{
						ClassLoader newTCC = managedBootstraps.get(bundleId).getClassLoader();
						Thread.currentThread().setContextClassLoader(newTCC);
						managedBootstraps.get(bundleId).getBootstrap().shutdown();  
						managedBootstraps.get(bundleId).getInstanceSR().unregister();
						managedBootstraps.remove(bundleId);
					}catch(RuntimeException e){
						e.printStackTrace();
					}finally{
						Thread.currentThread().setContextClassLoader(oldTCC);
					}	
				}
			}
		}

		@Override
		public void removedBundle(Bundle bundle, BundleEvent event,
				Object object) {	
		}	
    }
}
