/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.osgicdi.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;

import javax.enterprise.inject.Any;
import javax.enterprise.inject.Default;
import javax.enterprise.inject.Instance;
import javax.enterprise.util.Nonbinding;
import javax.inject.Qualifier;

import org.glassfish.osgicdi.Property;
import org.glassfish.osgicdi.Publish;
import org.glassfish.osgicdi.ServiceProperties;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleReference;

/**
 * This class obtains @Publish class instance of the current bean bundle and publishes
 * it as OSGi services all that are {@link Publish} annotated.
 * <p/>
 * @author Tang Yong (tangyong@cn.fujitsu.com)
 */
public class OSGiServicePublisher {

	private static Logger logger = Logger.getLogger(OSGiServicePublisher.class.getPackage().getName());
	
	private List<Class<?>> classes;
	
	private List<Instance<Object>> instanceList;
	
	public OSGiServicePublisher(List<Class<?>> classes, List<Instance<Object>> instanceList) {
		this.classes = classes;
		this.instanceList = instanceList;
	}
	
	public void RegisterOSGiService(){
		for(Class<?>clazz : classes){
			List<Annotation> qualifiers = getQualifiers(clazz);
			Object service = null;
			for (Instance<Object> instance : instanceList){
				try{
					 service = instance.select(clazz, qualifiers.toArray(new Annotation[qualifiers.size()])).get();
					 if (service != null){
						 break;
					 }
				}catch(Throwable t){
					t.printStackTrace();
				}
			}
								
			publish(clazz, service, qualifiers);
		}
	}
	
	private  List<Annotation> getQualifiers(Class<?> clazz) {
        List<Annotation> qualifiers = new ArrayList<Annotation>();
        for (Annotation annotation : clazz.getAnnotations()) {
            if (annotation.annotationType().isAnnotationPresent(Qualifier.class)) {
                qualifiers.add(annotation);
            }
        }
        return qualifiers;
    }
	
	/**
     * This method publish a bean class as an OSGi service.
     * <p/>
     * The class is published under (in order):
     * <ul>
     *  <li> The contract list of its @Publish annotation
     *  <li> Its interfaces
     *  <li> Its supertype
     *  <li> Itself implementation class
     * </ul>
     * with the right properties extracted from its annotations.
     * @param clazz
     * @param service
     * @param qualifiers
     */
    private void publish(Class<?> clazz, Object service, List<Annotation> qualifiers) {
        Publish publish = clazz.getAnnotation(Publish.class);
        Class<?>[] contracts = publish.contracts();
        
        Properties properties = getServiceProperties(qualifiers);
        
        BundleContext ctx = BundleReference.class.cast(clazz.getClassLoader())
                                  .getBundle().getBundleContext();
        
        if (contracts.length > 0) {
        	//if there are contracts and these contracts are valid
            String[] names = new String[contracts.length];
            for (int i = 0; i < contracts.length; i++) {
                if (contracts[i].isAssignableFrom(clazz)) {
                    names[i] = contracts[i].getName();                    
                } else {
                    RuntimeException e = new RuntimeException("Contract "
                            + contracts[i]
                            + " is not assignable from "
                            + clazz
                            + ", or is not an interface."
                            + " Unable to publish the service "
                            + clazz);
                   
                    throw e;
                }
            }
            
            ctx.registerService(names, service, properties);
        } else {
            if (clazz.getInterfaces().length > 0) {
            	//if there are multiple interfaces
                List<Class<?>> interfaces = new ArrayList<Class<?>>();
                for (Class<?> itf : clazz.getInterfaces()) {
                	interfaces.add(itf);
                }
                
                String[] names = new String[interfaces.size()];
                for (int i = 0; i < interfaces.size(); i++) {
                    names[i] = interfaces.get(i).getName();                   
                }
                
                ctx.registerService(names, service, properties);

            } else {
                Class<?> superClass = clazz.getClass().getSuperclass();
                if (superClass != null && superClass != Object.class) {
                	//if there is a superclass and the superclass is not equal Object
                    ctx.registerService(superClass.getName(), service, properties);
                } else {
                	// register the implementation type itself directly
                    ctx.registerService(clazz.getName(), service, properties);
                }
            }
        }
    }

	private Properties getServiceProperties(List<Annotation> qualifiers) {
		Properties properties = null;
		
		if (!qualifiers.isEmpty()) {
			properties = new Properties();
			
			for (Annotation qualifier : qualifiers) {
				if (qualifier.annotationType().equals(ServiceProperties.class)) {
                    for (Property property : ((ServiceProperties) qualifier).value()) {
                        properties.setProperty(property.name(), property.value());
                    }
                } else if (!qualifier.annotationType().equals(Default.class) && !qualifier.annotationType().equals(Any.class)) {
                    if (qualifier.annotationType().getDeclaredMethods().length > 0) {
                        for (Method m : qualifier.annotationType().getDeclaredMethods()) {
                            if (!m.isAnnotationPresent(Nonbinding.class)) {
                                try {
                                    String key = qualifier.annotationType().getSimpleName().toLowerCase()
                                            + "." + m.getName().toLowerCase();
                                    Object value = m.invoke(qualifier);
                                    if (value == null) {
                                        value = m.getDefaultValue();
                                        if (value == null) {
                                            value = "*";
                                        }
                                    }
                                    properties.setProperty(key, value.toString());
                                } catch (Throwable t) {// ignore
                                }
                            }
                        }
                    } else {
                        properties.setProperty(qualifier.annotationType().getSimpleName().toLowerCase(), "*");
                    }
                }
			}
		}
		
		return properties;
	}
}
